/* giFTcurs - curses interface to giFT
 * Copyright (C) 2001, 2002, 2003 Gran Weinholt <weinholt@dtek.chalmers.se>
 * Copyright (C) 2003 Christian Hggstrm <chm@c00.info>
 *
 * This file is part of giFTcurs.
 *
 * giFTcurs is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * giFTcurs is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with giFTcurs; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,  USA.
 *
 * $Id: transfer.c,v 1.112 2003/04/03 22:38:40 chnix Exp $
 */
#include "giftcurs.h"

#include <stdlib.h>
#include <ctype.h>
#include <assert.h>
#include <string.h>
#include <limits.h>

#include "parse.h"
#include "get.h"
#include "transfer.h"
#include "gift.h"
#include "screen.h"
#include "protocol.h"
#include "settings.h"
#include "format.h"
#include "poll.h"

/* This should never be used. It's only here so it can be translated. */
static const char *_statuses[] __attribute__ ((unused));
static const char *_statuses[] = {
	N_("Waiting"),
	N_("Paused"),
	N_("Queued (Remotely)"),
	N_("Queued"),
	N_("Active"),
	N_("Complete"),
	N_("Cancelled (Remotely)"),
	N_("Timed out"),
	N_("Awaiting connection"),
	N_("Connecting"),
	N_("Receiving"),
	N_("Verifying"),
	/*N_("Queued (position %s)"), */
	N_("Sending"),
	N_("Received HTTP PUSH"),
	N_("Received HTTP headers"),
	N_("Sent HTTP request"),
	N_("Malformed HTTP header"),
	N_("Remote host had an aneurism"),
	N_("Connection closed"),
	N_("Connection refused"),
	N_("File not found"),
	N_("Local read error"),
	N_("Incorrect chunk size"),
	N_("No route to host"),
	/* XXX: this list will hopefully be obsolete when giFT gets i18n */
};

#define BANDWIDTH_TIMER SECS(2)

int autodestroy_all = 0;
tree downloads;
tree uploads;

static int nr_alive = 0;
static list download_sort_methods = LIST_INITIALIZER;
static list upload_sort_methods = LIST_INITIALIZER;
static const char default_sort_methods[] = "filesize;transferred;bandwidth;filename";
static int upload_sort_method, download_sort_method;

static void transfer_change_handle(ntree * packet, void *udata);
static void transfer_add_handle(ntree * packet, void *udata);
static void addsource_parse(ntree * packet, transfer *);
static tick_t bandwidth_timer(void *udata);
static void estimate_speed(struct bandwidth_meas *bw, int sizediff, tick_t dt, tick_t idle_time);
static void transfer_detach_all(ntree *, void *udata);

void transfers_init(void)
{
	char *methods;

	tree_initialize(&downloads);
	tree_initialize(&uploads);

	autodestroy_all = atoi(get_config("set", "autoclear", "0"));

	methods = get_config("sort", "downloads", NULL);
	if (!methods)
		methods = get_config("sort", "transfers", (char *) default_sort_methods);
	split(methods, ';', &download_sort_methods);

	methods = get_config("sort", "uploads", NULL);
	if (!methods)
		methods = get_config("sort", "transfers", (char *) default_sort_methods);
	split(methods, ';', &upload_sort_methods);

	upload_sort_method = download_sort_method = 0;

	gift_register("ADDDOWNLOAD", (EventCallback) transfer_add_handle, (void *) TRANSFER_DOWNLOAD);
	gift_register("ADDUPLOAD", (EventCallback) transfer_add_handle, (void *) TRANSFER_UPLOAD);
	gift_register("DETACH", (EventCallback) transfer_detach_all, NULL);

	methods = list_index(&upload_sort_methods, 0);
	tree_sort(&uploads, methods);
	methods = list_index(&download_sort_methods, 0);
	tree_sort(&downloads, methods);
}

void source_destroy(source * s)
{
	free(s->user);
	free(s->node);
	free(s->href);
	free(s->status);
	tree_node_destroy(s);
}

void transfer_destroy(transfer * t)
{
	transfer_detach(t);
	free(t->protocol);
	free(t->filename);
	free(t->hash);
	free(t->status);
	tree_node_destroy(t);
}

/* Replace a allocated string with another.
 * Avoids malloc/free if the strings are equal. */
static void change_string(char **dst, const char *src)
{
	if (!*dst) {
		*dst = strdup(src);
	} else if (strcmp(*dst, src)) {
		free(*dst);
		*dst = strdup(src);
	}
}

void source_detach(source * s)
{
	/* What is an appropriate status here? */
	change_string(&s->status, N_("Inactive"));
}

void transfer_detach(transfer * t)
{
	t->paused = 0;

	if (t->id) {
		gift_unregister_id(t->id);
		t->id = 0;

		if (--nr_alive == 0)
			poll_del_timer(bandwidth_timer);
	}

	/*list_destroy_all(&t->sourcen); */
	list_foreach(tree_children(t), (LFunc) source_detach);

	if (t->search_id) {
		gift_search_stop(t->search_id, "LOCATE");
		t->search_id = 0;
	}
}

static void transfer_detach_panic(transfer * t)
{
	/* This is called when giFT disconnects without a word. */
	int pos = tree_hold(t);

	if (t->id)
		change_string(&t->status, N_("Unknown status"));
	transfer_detach(t);

	tree_release(t, pos);
}

static void transfer_detach_all(ntree * unused, void *udata)
{
	list_foreach(tree_children(&downloads), (LFunc) transfer_detach_panic);
	list_foreach(tree_children(&uploads), (LFunc) transfer_detach_panic);
	if (downloads.update_ui)
		downloads.update_ui();
	if (uploads.update_ui)
		uploads.update_ui();
}

void transfer_forget(transfer * t)
{
	tree_remove(t);
	transfer_destroy(t);
}

static transfer *transfer_new(gift_id id, int type, char *name, unsigned int size)
{
	transfer *t = NEW_NODE(transfer);

	t->id = id;
	t->filename = strdup(name);
	t->filesize = size;
	t->type = type;
	t->protocol = NULL;
	list_initialize(&t->tnode.children);
	transfer_reset(t);
	return t;
}

void transfer_cancel(transfer * t)
{
	ntree *packet = NULL;

	if (!t->id)
		return;

	interface_append_int(&packet, "TRANSFER", t->id);
	interface_append(&packet, "action", "cancel");
	gift_write(&packet);
}

void source_cancel(transfer * t, source * s)
{
	ntree *packet = NULL;

	if (!t->id)
		return;

	interface_append_int(&packet, "DELSOURCE", t->id);
	interface_append(&packet, "url", s->href);
	gift_write(&packet);
}

void transfer_reset(transfer * t)
{
	/* reset progress statistics */
	t->transferred = 0;
	memset(&t->bw, 0, sizeof t->bw);
}

void transfer_suspend(transfer * t)
{
	ntree *packet = NULL;

	if (!t->id)
		return;

	interface_append_int(&packet, "TRANSFER", t->id);
	interface_append(&packet, "action", t->paused ? "unpause" : "pause");
	gift_write(&packet);
}

/* Find a (dead) download suitable for reuse */
transfer *transfer_find(tree * tr, char *hash, unsigned int size, int dead)
{
	int i;

	if (!hash)
		return NULL;

	for (i = 0; i < tree_children(tr)->num; i++) {
		transfer *t = list_index(tree_children(tr), i);

		if (dead && t->id)
			continue;
		if (t->filesize == size && t->hash && !strcmp(t->hash, hash))
			return t;
	}
	return NULL;
}

transfer *transfer_find_id(tree * tr, gift_id id)
{
	int i;

	for (i = 0; i < tree_children(tr)->num; i++) {
		transfer *t = list_index(tree_children(tr), i);

		if (t->id == id)
			return t;
	}
	return NULL;
}

int transfer_alive(transfer * t)
{
	return !!t->id;
}

source *lookup_source(transfer * t, const char *user, const char *href)
{
	int i;

	for (i = 0; i < tree_children(t)->num; i++) {
		source *s = list_index(tree_children(t), i);

		if (!strcmp(user, s->user) && !strcmp(href, s->href))
			return s;
	}

	return NULL;
}

static void source_change_parse(ntree * param, transfer * t)
{
	char *user, *href, *status;
	source *s;
	unsigned int transmit, total, start;
	int diff;
	int pos;

	if (strcmp(interface_name(param), "SOURCE"))
		return;

	user = interface_lookup(param, "user");
	href = interface_lookup(param, "url");
	/* FIXME: use status if there is no action. I hope this is only for
	   backwards compatibility. */
	if (!(status = interface_lookup(param, "action")))
		status = interface_lookup(param, "status");

	if (!user || !href || !status) {
		DEBUG("Bad SOURCE.");
		return;
	}

	if (!(s = lookup_source(t, user, href))) {
		DEBUG("A source I didn't know about was changed");
		return;
	}

	pos = tree_hold(s);

	change_string(&s->status, status);

	total = my_atoi(interface_lookup(param, "total"));
	transmit = my_atoi(interface_lookup(param, "transmit"));
	start = my_atoi(interface_lookup(param, "start"));

	/* total(), transmit() and start() is skipped when status(Paused) */
	/* Also, don't measure bandwidth on the first transmit() tag */
	if (total && s->total) {
		/* Check if this source has started on a new chunk */
		if (start != s->start)
			diff = transmit + (s->total - s->transmit);
		else
			diff = transmit - s->transmit;

		if (diff < 0) {
			DEBUG("Hmm.. negative bandwidth");
			diff = 0;
		}

		estimate_speed(&s->bw, diff, t->elapsed, 0);
	} else {
		memset(&s->bw, 0, sizeof s->bw);
	}

	s->transmit = transmit;
	s->start = start;
	s->total = total;

	tree_release(s, pos);
}

static void addsource_parse(ntree * param, transfer * t)
{
	char *user, *href, *status;
	source *s;

	user = interface_lookup(param, "user");
	href = interface_lookup(param, "url");
	status = interface_lookup(param, "status");

	if (!user || !href || !status) {
		DEBUG("Bad SOURCE.");
		return;
	}

	if (lookup_source(t, user, href)) {
		DEBUG("Already had that source");
		return;
	}
	s = NEW_NODE(source);
	s->user = strdup(user);
	s->node = my_strdup(interface_lookup(param, "node"));
	s->href = strdup(href);

	change_string(&s->status, status);
	s->start = my_atoi(interface_lookup(param, "start"));
	s->transmit = my_atoi(interface_lookup(param, "transmit"));
	s->total = my_atoi(interface_lookup(param, "total"));

	if (!t->protocol && strchr(href, ':'))
		t->protocol = strndup(href, strchr(href, ':') - href);
	tree_insort(t, s);
}

static void transfer_update_fields(transfer * t, ntree * param)
{
	char *state = interface_lookup(param, "state");

	t->transferred = my_atoi(interface_lookup(param, "transmit"));

	if (!state)
		return;

	if (!t->status || strcmp(state, t->status)) {
		free(t->status);
		t->status = strdup(state);
		t->paused = !strcasecmp(state, "paused");
		if (t->paused)
			memset(&t->bw, 0, sizeof t->bw);
	}
}

static void transfer_add_handle(ntree * param, void *udata)
{
	gift_id id;
	tree *transfer_tree;
	int transfer_type = (int) udata;
	transfer *t;
	char *filename;
	char *hash;
	unsigned int size;
	int pos = -1;

	id = my_atoi(interface_value(param));

	if (!id) {
		DEBUG("Daemon sent a spooky transfer with a zero id.");
		return;
	}

	transfer_tree = transfer_type == TRANSFER_DOWNLOAD ? &downloads : &uploads;

	if (transfer_find_id(transfer_tree, id)) {
		DEBUG("Got transfer_ADD but already have it");
		return;
	}
	filename = interface_lookup(param, "file");
	hash = interface_lookup(param, "hash");
	size = my_atoi(interface_lookup(param, "size"));

	if (!size)
		return;

	if (!filename)
		filename = "";

	/* Check if we can reuse an existing (lost) transfer */
	/* Maybe we should search by name here */
	if ((t = transfer_find(transfer_tree, hash, size, 1))) {
		pos = tree_hold(t);
		free(t->filename);
		t->filename = strdup(filename);
		t->id = id;
	} else {
		t = transfer_new(id, transfer_type, filename, size);
		t->hash = my_strdup(hash);
	}

	transfer_update_fields(t, param);

	if (transfer_type == TRANSFER_UPLOAD) {
		if (!my_atoi(interface_lookup(param, "shared")))
			t->autodestroy = 1;
	}

	interface_foreach(param, (PForEachFunc) addsource_parse, t);

	if (pos == -1)
		tree_insort(transfer_tree, t);
	else
		tree_release(t, pos);

	if (transfer_tree->update_ui)
		transfer_tree->update_ui();

	if (nr_alive++ == 0)
		poll_add_timer(BANDWIDTH_TIMER, bandwidth_timer, NULL);

	gift_register_id(id, (EventCallback) transfer_change_handle, t);
}

static void transfer_change_handle(ntree * param, void *udata)
{
	tree *transfer_tree;
	int transfer_type;
	transfer *t = udata;
	char *command = interface_name(param);
	int pos;

	transfer_type = t->type;
	transfer_tree = t->tnode.parent;

	t->last_time = uptime();

	if (!strcmp(command, "ADDSOURCE")) {
		pos = tree_hold(t);
		addsource_parse(param, t);
		tree_release(t, pos);
	} else if (!strcmp(command, "DELSOURCE")) {
		char *user, *href, *status;
		source *s;

		user = interface_lookup(param, "user");
		href = interface_lookup(param, "url");
		status = interface_lookup(param, "status");

		if (!user || !href || !status) {
			DEBUG("Bad SOURCE.");
			return;
		}

		if (!(s = lookup_source(t, user, href))) {
			DEBUG("Didn't have that source");
			return;
		}

		pos = tree_hold(t);

		tree_remove(s);
		source_destroy(s);

		tree_release(t, pos);
	} else if (strstr(command, "DEL")) {
		if (transfer_type == TRANSFER_DOWNLOAD)
			message(t->transferred == t->filesize ? _("Download of %s completed.") :
					_("Download of %s cancelled."), t->filename);

		if (autodestroy_all || t->autodestroy) {
			transfer_forget(t);
		} else {
			pos = tree_hold(t);
			transfer_detach(t);
			if (t->transferred < t->filesize)
				change_string(&t->status, N_("Cancelled"));
			tree_release(t, pos);
		}
	} else {					/* CHG* */
		unsigned int throughput, elapsed;

		pos = tree_hold(t);

		transfer_update_fields(t, param);

		throughput = my_atoi(interface_lookup(param, "throughput"));
		elapsed = my_atoi(interface_lookup(param, "elapsed"));
		t->elapsed = elapsed = (SECS(elapsed + 5)) / 1000;

		estimate_speed(&t->bw, throughput, elapsed, 0);

		/* source_change_parse wants t->elapsed to be set */
		interface_foreach(param, (PForEachFunc) source_change_parse, t);

		tree_release(t, pos);
	}

	if (transfer_tree->update_ui)
		transfer_tree->update_ui();
}

#define ESTIMATE_TIME SECS(8)

/* This function recalculates the bandwidth, taking account to history of
 * previous samples, a new sample, and the time this transfer has been idle.
 */
static void estimate_speed(struct bandwidth_meas *bw, int transferred, tick_t elapsed,
						   tick_t total_time)
{
	int i;
	unsigned int total_size;

	//DEBUG("%p: transferred=%d elapsed=%d idle=%d", bw, transferred, elapsed, total_time);

	/* check if deltatime is enough */
	if (bw->samples[0].dt + elapsed < ESTIMATE_TIME / buflen(bw->samples)) {
		/* add this onto the previous one */
		bw->samples[0].dt += elapsed;
		bw->samples[0].dsize += transferred;
	} else {
		/* shift all previous samples down */
		memmove(bw->samples + 1, bw->samples, sizeof bw->samples[0] * (buflen(bw->samples) - 1));
		bw->samples[0].dt = elapsed;
		bw->samples[0].dsize = transferred;
	}

	total_size = 0;

	for (i = 0; i < buflen(bw->samples) && total_time < ESTIMATE_TIME; i++) {
		unsigned int dt = bw->samples[i].dt, dsize = bw->samples[i].dsize;

		//DEBUG("%p: dt=%d ds=%d", bw, dt, dsize);

		if (total_time + dt > ESTIMATE_TIME) {
			total_size += (ESTIMATE_TIME - total_time) * dsize / dt;
			total_time = ESTIMATE_TIME;
			break;
		}
		total_time += dt;
		total_size += dsize;
	}
	if (total_size == 0)
		bw->bandwidth = 0;
	else if (total_time == 0)
		bw->bandwidth = INT_MAX;	/* prevent division by zero */
	else
		bw->bandwidth = SECS(total_size) / total_time;

	//DEBUG("%p: size=%d time=%d bandwidth=%d", bw, total_size, total_time, bw->bandwidth);
}

static int transfer_calculate(transfer * t)
{
	int i;
	tick_t idle_time;
	int key, key2;

	if (t->paused)
		return 0;

	idle_time = uptime() - t->last_time;

	if (idle_time < BANDWIDTH_TIMER)
		return 0;

	key = tree_hold(t);

	/* Transfer is stalled. Start cutting down the bandwidth */
	estimate_speed(&t->bw, 0, 0, idle_time);

	/* Ditto for the sources */
	for (i = 0; i < tree_children(t)->num; i++) {
		source *s = list_index(tree_children(t), i);

		key2 = tree_hold(s);
		estimate_speed(&s->bw, 0, 0, idle_time);
		tree_release(s, key2);
	}

	tree_release(t, key);
	return 1;
}

static tick_t bandwidth_timer(void *udata)
{
	int changed, i;

	for (changed = i = 0; i < tree_children(&downloads)->num; i++)
		changed += transfer_calculate(list_index(tree_children(&downloads), i));
	if (changed && downloads.update_ui)
		downloads.update_ui();

	for (changed = i = 0; i < tree_children(&uploads)->num; i++)
		changed += transfer_calculate(list_index(tree_children(&uploads), i));
	if (changed && uploads.update_ui)
		uploads.update_ui();
	return BANDWIDTH_TIMER;
}

char *transfer_change_sort_method(tree * active_tree, int dir)
{
	/* direction == 0 just sorts with current method. */
	/* direction == -1 or +1 changes method and sorts */
	int i;
	char *s;

	list *sorting_methods = active_tree == &uploads ? &upload_sort_methods : &download_sort_methods;
	int *sort_index = active_tree == &uploads ? &upload_sort_method : &download_sort_method;

	/* Change sorting method */
	i = *sort_index;
	assert(i >= 0);

	i += dir;
	if (i < 0)
		i = sorting_methods->num - 1;
	else if (i >= sorting_methods->num)
		i = 0;

	*sort_index = i;

	s = list_index(sorting_methods, i);

	tree_sort(active_tree, s);

	return s;
}
